__version__ = "0.1.0"

records = {
    "经济":
        {
            "供给": {

            },
            "需求": {

            }

        },
    "通胀":
        {
            "食品分项":
                {
                    "猪肉":
                        {
                            "2018-08-24": " 农业农村部新闻办公室发布通知称，辽宁省沈阳市沈北新区发生一起生猪非洲猪瘟疫情，这是我国首次发生非洲猪瘟疫情。"
                        },
                    "蔬菜":
                        {
                            "2018-08-23": "受台风影响，中国菜都山东寿光遭遇水灾，蔬菜价格或存涨价预期。"
                        }

                }

        },
    "资金":
        {

        },
    "政策":
        {
            "2018-08-28": "财政部依法管控地方政府债务风险（新华社).",
            "2018-08-30": "国务院：对境外机构免征3年境内债券市场利息所得税"

        },
    "海外":
        {

        }
}
